
# Imports
import json
import os
import shutil
from pathlib import Path
from typing import Any, cast

import stouputils as stp
from beet import Font, Texture
from beet.core.utils import JsonDict, TextComponent
from PIL import Image

from stewbeet.core.definitions_helper.completion import add_private_custom_data_for_namespace

from ...core.__memory__ import Mem
from ...core.cls.block import Block, VanillaBlock
from ...core.cls.ingredients import CRAFTING_RECIPES_TYPES, Ingr
from ...core.cls.item import Item
from ...core.cls.recipe import (
	AwakenedForgeRecipe,
	BlastingRecipe,
	CampfireCookingRecipe,
	CraftingShapedRecipe,
	CraftingShapelessRecipe,
	PulverizingRecipe,
	SmeltingRecipe,
	SmithingTransformRecipe,
	SmithingTrimRecipe,
	SmokingRecipe,
	StonecuttingRecipe,
)
from ...core.cls.wiki_button import WikiButton
from ...core.constants import (
	CATEGORY,
	NO_SILK_TOUCH_DROP,
	OFFICIAL_LIBS,
	WIKI_COMPONENT,
)
from ...core.definitions_helper import add_item_name_and_lore_if_missing
from ...core.utils.io import convert_to_serializable, super_merge_dict, write_load_file
from ...core.utils.text_component import item_id_to_name
from ..custom_recipes.vanilla import VanillaRecipeHandler
from ..initialize.source_lore_font import find_pack_png
from ..resource_pack.item_models import AutoModel  # Handle new items models (used for the manual and the heavy workbench)
from .book_components import get_item_component
from .book_optimizer import optimize_element, remove_events
from .craft_content import generate_craft_content
from .dialog import generate_dialogs
from .image_utils import (
	add_border,
	careful_resize,
	generate_high_res_font,
	load_simple_case_no_border,
)
from .iso_renders import generate_all_iso_renders
from .other_utils import (
	convert_shapeless_to_shaped,
	generate_otherside_crafts,
	remove_duplicate_furnace_crafts,
	remove_unknown_crafts,
)
from .page_font import generate_page_font, generate_wiki_font_for_ingr
from .shared_import import (
	AWAKENED_3X3_FONT,
	AWAKENED_3X4_FONT,
	AWAKENED_FORGE_STRUCT_FONT,
	BOOK_FONT,
	BORDER_SIZE,
	FONT_FILE,
	FURNACE_FONT,
	HEAVY_WORKBENCH_CATEGORY,
	HOVER_AWAKENED_3X3_FONT,
	HOVER_AWAKENED_3X4_FONT,
	HOVER_EQUIVALENTS,
	HOVER_FURNACE_FONT,
	HOVER_MINING_FONT,
	HOVER_PULVERIZING_FONT,
	HOVER_SHAPED_2X2_FONT,
	HOVER_SHAPED_3X3_FONT,
	HOVER_STONECUTTING_FONT,
	INVISIBLE_ITEM_FONT,
	MANUAL_ASSETS_PATH,
	MEDIUM_NONE_FONT,
	MICRO_NONE_FONT,
	MINING_FONT,
	NONE_FONT,
	PULVERIZING_FONT,
	SHAPED_2X2_FONT,
	SHAPED_3X3_FONT,
	SMALL_NONE_FONT,
	STONECUTTING_FONT,
	TEMPLATES_PATH,
	VERY_SMALL_NONE_FONT,
	WIKI_INFO_FONT,
	WIKI_INGR_OF_CRAFT_FONT,
	WIKI_NONE_FONT,
	WIKI_RESULT_OF_CRAFT_FONT,
	SharedMemory,
	get_border_color,
	get_next_font,
	get_page_font,
	get_page_number,
)
from .showcase_image import generate_showcase_images
from .stardust_forge import get_stardust_forge_page


# Utility functions
def deepcopy(x: Any) -> Any:
	""" Deep copy using JSON serialization, converting WikiButton objects. """
	return json.loads(json.dumps(convert_to_serializable(x)))

def manual_main():
	# Copy everything in the manual assets folder to the templates folder
	with stp.super_open(f"{TEMPLATES_PATH}/.gitignore", "w") as f:
		f.write("*")
	shutil.copytree(MANUAL_ASSETS_PATH + "/assets", TEMPLATES_PATH, dirs_exist_ok = True)

	# Copy the manual_overrides folder to the templates folder
	manual_overrides: str = Mem.ctx.meta.get("stewbeet",{}).get("manual", {}).get("manual_overrides", "")
	if manual_overrides and os.path.exists(manual_overrides):
		shutil.copytree(manual_overrides, TEMPLATES_PATH, dirs_exist_ok = True)
		with stp.super_open(f"{TEMPLATES_PATH}/.gitignore", "w") as f:
			f.write("*") # Ensure the .gitignore file is present to avoid committing manual overrides

	# Launch the routine
	routine()

def routine():
	manual_config: JsonDict = Mem.ctx.meta.get("stewbeet",{}).get("manual", {})
	json_dump_path: str = manual_config.get("json_dump_path", "")
	manual_name: str = manual_config.get("name", "")
	if not manual_name:
		manual_name = f"{Mem.ctx.project_name} Manual"
	if len(manual_name) >= 32:
		stp.error(f"Manual name '{manual_name}' is too long (max 32 characters), Minecraft does not support it. Please change it in the stewbeet config.")

	# If smithed crafter is used, add it to the manual (last page that we will move to the second page)
	if OFFICIAL_LIBS["smithed.crafter"]["is_used"]:
		Mem.ctx.assets[Mem.ctx.project_id].textures["item/heavy_workbench"] = Texture(source_path=f"{TEMPLATES_PATH}/heavy_workbench.png")
		obj = Block(
			id="heavy_workbench",
			vanilla_block=VanillaBlock(id=""),
			manual_category=HEAVY_WORKBENCH_CATEGORY,
			override_model={
				"parent":"minecraft:block/cube",
				"texture_size":[64,32],
				"textures":{"0":f"{Mem.ctx.project_id}:item/heavy_workbench"},
				"elements":[{"from":[0,0,0],"to":[16,16,16],"faces":{"north":{"uv":[4,8,8,16],"texture":"#0"},"east":{"uv":[0,8,4,16],"texture":"#0"},"south":{"uv":[12,8,16,16],"texture":"#0"},"west":{"uv":[8,8,12,16],"texture":"#0"},"up":{"uv":[4,0,8,8],"texture":"#0"},"down":{"uv":[8,0,12,8],"texture":"#0"}}}],
				"display":{"thirdperson_righthand":{"rotation":[75,45,0],"translation":[0,2.5,0],"scale":[0.375,0.375,0.375]},"thirdperson_lefthand":{"rotation":[75,45,0],"translation":[0,2.5,0],"scale":[0.375,0.375,0.375]},"firstperson_righthand":{"rotation":[0,45,0],"scale":[0.4,0.4,0.4]},"firstperson_lefthand":{"rotation":[0,225,0],"scale":[0.4,0.4,0.4]},"ground":{"translation":[0,3,0],"scale":[0.25,0.25,0.25]},"gui":{"rotation":[30,225,0],"scale":[0.625,0.625,0.625]},"head":{"translation":[0,-30.43,0],"scale":[1.601,1.601,1.601]},"fixed":{"scale":[0.5,0.5,0.5]}}
			},
			recipes=[
				CraftingShapedRecipe(shape=["###","#C#","SSS"],ingredients={"#":Ingr("minecraft:oak_log"),"C":Ingr("minecraft:crafting_table"),"S":Ingr("minecraft:smooth_stone")})
			],
			components={
				"item_name": "Heavy Workbench",
				"item_model": f"{Mem.ctx.project_id}:heavy_workbench",
			},
		)
		AutoModel.from_definitions(obj, {}, ignore_textures = True).process()

	# Prework
	os.makedirs(f"{SharedMemory.cache_path}/font/page", exist_ok=True)
	os.makedirs(f"{SharedMemory.cache_path}/font/wiki_icons", exist_ok = True)
	os.makedirs(f"{SharedMemory.cache_path}/font/high_res", exist_ok = True)
	generate_all_iso_renders()

	# Check if there is any awakened forge recipe with 3x3 or 3x4 ingredients
	has_forge_3x3: bool = False
	has_forge_3x4: bool = False
	for item in Mem.definitions.keys():
		obj = Item.from_id(item)
		for recipe in obj.recipes:
			if recipe.get("type") == AwakenedForgeRecipe.type:
				if len(recipe["ingredients"]) <= 9:
					has_forge_3x3 = True
				else:
					has_forge_3x4 = True
		if has_forge_3x3 and has_forge_3x4:
			break

	# Constants
	FONT = Mem.ctx.project_id + ':' + FONT_FILE
	MAX_ITEMS_PER_ROW: int = min(6, manual_config.get("max_items_per_row", 5))
	MAX_ROWS_PER_PAGE: int = min(7, manual_config.get("max_rows_per_page", 5))
	MAX_ITEMS_PER_PAGE = MAX_ITEMS_PER_ROW * MAX_ROWS_PER_PAGE # (for showing up all items in the categories pages)

	# Calculate left padding for categories pages depending on config['max_items_per_row']: higher the value, lower the padding
	LEFT_PADDING = 6 - MAX_ITEMS_PER_ROW
	# Copy assets in the resource pack
	if not manual_config.get("debug_mode", False):
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/none"] = Texture(source_path=f"{TEMPLATES_PATH}/none_release.png")
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/invisible_item"] = Texture(source_path=f"{TEMPLATES_PATH}/invisible_item_release.png")
	else:
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/none"] = Texture(source_path=f"{TEMPLATES_PATH}/none.png")
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/invisible_item"] = Texture(source_path=f"{TEMPLATES_PATH}/invisible_item.png")
	Mem.ctx.assets[Mem.ctx.project_id].textures["font/wiki_information"] = Texture(source_path=f"{TEMPLATES_PATH}/wiki_information.png")
	Mem.ctx.assets[Mem.ctx.project_id].textures["font/wiki_result_of_craft"] = Texture(source_path=f"{TEMPLATES_PATH}/wiki_result_of_craft.png")
	Mem.ctx.assets[Mem.ctx.project_id].textures["font/wiki_ingredient_of_craft"] = Texture(source_path=f"{TEMPLATES_PATH}/wiki_ingredient_of_craft.png")
	if SharedMemory.high_resolution:
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/shaped_2x2"] = Texture(source_path=f"{TEMPLATES_PATH}/shaped_2x2.png")
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/shaped_3x3"] = Texture(source_path=f"{TEMPLATES_PATH}/shaped_3x3.png")
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/furnace"] = Texture(source_path=f"{TEMPLATES_PATH}/furnace.png")
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/stonecutting"] = Texture(source_path=f"{TEMPLATES_PATH}/stonecutting.png")
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/pulverizing"] = Texture(source_path=f"{TEMPLATES_PATH}/pulverizing.png")
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/mining"] = Texture(source_path=f"{TEMPLATES_PATH}/mining.png")
		if has_forge_3x3:
			Mem.ctx.assets[Mem.ctx.project_id].textures["font/awakened_forge_3x3"] = Texture(source_path=f"{TEMPLATES_PATH}/awakened_forge_3x3.png")
		if has_forge_3x4:
			Mem.ctx.assets[Mem.ctx.project_id].textures["font/awakened_forge_3x4"] = Texture(source_path=f"{TEMPLATES_PATH}/awakened_forge_3x4.png")
		if has_forge_3x3 or has_forge_3x4:
			Mem.ctx.assets[Mem.ctx.project_id].textures["font/awakened_forge_1"] = Texture(source_path=f"{TEMPLATES_PATH}/awakened_forge_1.png")
			Mem.ctx.assets[Mem.ctx.project_id].textures["font/awakened_forge_2"] = Texture(source_path=f"{TEMPLATES_PATH}/awakened_forge_2.png")
	if SharedMemory.use_dialog > 0:
		Mem.ctx.assets[Mem.ctx.project_id].textures["font/book"] = Texture(source_path=f"{TEMPLATES_PATH}/book.png")

	# Prepare category padding if dialog
	category_padding: list[str] = [VERY_SMALL_NONE_FONT] if SharedMemory.use_dialog > 0 else []

	# Generate categories list
	categories: dict[str, list[str]] = {}
	for item in Mem.definitions.keys():
		obj = Item.from_id(item)

		if not obj.manual_category:
			stp.suggestion(f"Item '{item}' has no category key. Skipping.")
			continue

		if obj.manual_category not in categories:
			categories[obj.manual_category] = []
		categories[obj.manual_category].append(item)

	# Error message if there is too many categories
	if len(categories) > MAX_ITEMS_PER_PAGE:
		stp.error(f"Too many categories ({len(categories)}). Maximum is {MAX_ITEMS_PER_PAGE}. Please reduce the number of item categories.")

	# Debug categories and sizes
	s = ""
	for file, items in categories.items():
		if file == HEAVY_WORKBENCH_CATEGORY:
			continue
		s += f"\n- {file}: {len(items)} items"
		if len(items) > MAX_ITEMS_PER_PAGE:
			s += f" (splitted into {len(items) // MAX_ITEMS_PER_PAGE + 1} pages)"
	nb_categories: int = len(categories) - (1 if HEAVY_WORKBENCH_CATEGORY in categories else 0)
	stp.debug(f"Found {nb_categories} categories in the definitions:{s}")

	# Split up categories into pages
	categories_pages: dict[str, list[str]] = {}
	for file, items in categories.items():
		if file != HEAVY_WORKBENCH_CATEGORY:
			i = 0
			while i < len(items):
				page_name = file.title()
				if len(items) > MAX_ITEMS_PER_PAGE:
					number = i // MAX_ITEMS_PER_PAGE + 1
					page_name += f" #{number}"
				new_items = items[i:i + MAX_ITEMS_PER_PAGE]
				categories_pages[page_name] = new_items
				i += MAX_ITEMS_PER_PAGE

	# Sort items depending on their category order
	items_with_category = [x for x in [(item, Item.from_id(item)) for item in Mem.definitions.keys()] if x[1].manual_category]
	category_list = list(categories.keys())
	sorted_definitions_on_category = sorted(items_with_category, key = lambda x: category_list.index(x[1].manual_category or ""))

	# If the manual cache is enabled and we have a cache file, load it
	cache_pages: bool = manual_config.get("cache_pages", False)
	if cache_pages and json_dump_path and os.path.exists(json_dump_path) and os.path.exists(f"{SharedMemory.cache_path}/font/manual.json"):
		with stp.super_open(json_dump_path, "r") as f:
			book_content: list[list[TextComponent]] = json.load(f)

		# Make the page dictionary for later use (dialog)
		i: int = 2
		if OFFICIAL_LIBS["smithed.crafter"]["is_used"]:
			SharedMemory.manual_pages.append({"number": i, "name": "heavy_workbench"})
			i += 1
		if has_forge_3x3 or has_forge_3x4:
			SharedMemory.manual_pages.append({"number": i, "name": "stardust_forge"})
			i += 1
		i += 1 # Skip page linking to each category
		for page_name in categories_pages.keys():
			SharedMemory.manual_pages.append({"number": i, "name": page_name})
			i += 1
		for item, _ in sorted_definitions_on_category:
			SharedMemory.manual_pages.append({"number": i, "name": item})
			i += 1
	# Else, generate all
	else:
		## Prepare pages (append categories first, then items depending on categories order)
		i = 2 # Skip first two pages (introduction + categories)

		# Append categories
		for page_name, items in categories_pages.items():
			i += 1
			SharedMemory.manual_pages.append({"number": i, "name": page_name, "raw_data": items, "type": CATEGORY})

		# Append items (sorted by category)
		for item, data in sorted_definitions_on_category:
			i += 1
			SharedMemory.manual_pages.append({"number": i, "name": item, "raw_data": data, "type": "item"})

		# Prepare definitions as Item objects
		definitions_as_objects: dict[str, Item] = {item: Item.from_id(item) for item in Mem.definitions.keys()}

		# Encode pages
		book_content: list[list[TextComponent]] = []
		os.makedirs(f"{SharedMemory.cache_path}/font/category", exist_ok=True)
		simple_case = load_simple_case_no_border(SharedMemory.high_resolution)	# Load the simple case image for later use in categories pages
		def encode_page(page: JsonDict):
			content: list[TextComponent] = []
			number = page["number"]
			raw_data: Item | list[str] = page["raw_data"]
			page_font = ""
			if not SharedMemory.high_resolution:
				page_font = get_page_font(number)
			name = str(page["name"])
			titled = item_id_to_name(name) + "\n"

			# Encode categories {'number': 2, 'name': 'Material #1', 'raw_data': ['adamantium_block', 'adamantium_fragment', ...]}
			if page["type"] == CATEGORY and isinstance(raw_data, list):
				file_name = name.replace(" ", "_").replace("#", "").lower()
				page_font = get_page_font(number)
				SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/category/{file_name}.png", "ascent": 1, "height": 131, "chars": [page_font]})
				content.append({"text": "", "font": FONT, "color": "white"})	# Make default font for every next component
				content.append({"text": "➤ ", "font": "minecraft:default", "color": "black"})
				content.append({"text": titled, "font": "minecraft:default", "color": "black", "underlined": True})
				content.append(SMALL_NONE_FONT * LEFT_PADDING + page_font + "\n")

				# Prepare image and line list
				page_image = Image.new("RGBA", (256, 256), (0, 0, 0, 0))
				x, y = 2, 2	# Prevision for global border and implicit border
				line: list[TextComponent] = []

				# For each item in the category, get its page number and texture, then add it to the image
				max_items_reached: bool = False
				for item in raw_data:

					# Get item texture
					texture_path = f"{SharedMemory.cache_path}/items/{Mem.ctx.project_id}/{item}.png"
					if os.path.exists(texture_path):
						item_image = Image.open(texture_path)
					else:
						stp.warning(f"Missing texture at '{texture_path}', using empty texture")
						item_image = Image.new("RGBA", (1, 1), (0, 0, 0, 0))
					if not SharedMemory.high_resolution:
						resized = careful_resize(item_image, 32)
						high_res_font = None
					else:
						resized = Image.new("RGBA", (1, 1), (0, 0, 0, 0))	# Empty texture to use for category page
						high_res_font = generate_high_res_font(item, item_image)

					# Paste the simple case and the item_image
					page_image.paste(simple_case, (x, y))
					mask = resized.convert("RGBA").split()[3]
					page_image.paste(resized, (x + 2, y + 2), mask)
					x += simple_case.size[0]

					# Add the click_event part to the line and add the 2 times the line if enough items
					component = get_item_component(item, only_those_components=[x for x in SharedMemory.components_to_include if x != "lore"])
					component["text"] = MEDIUM_NONE_FONT if not SharedMemory.high_resolution else high_res_font
					line.append(component)
					if len(line) == MAX_ITEMS_PER_ROW:
						max_items_reached = True
						line.insert(0, SMALL_NONE_FONT * LEFT_PADDING)
						content.extend(deepcopy(line))
						content.extend(category_padding)
						for i in range(1, len(line)):
							selected = line[-i]
							if isinstance(selected, dict):
								selected["text"] = MEDIUM_NONE_FONT
						content.extend(["\n", *line, *category_padding, "\n"])
						line = []
						x = 2
						y += simple_case.size[1]

				# If remaining items in the line, add them
				if len(line) > 0:
					if SharedMemory.use_dialog > 0 and max_items_reached:
						line.append(MEDIUM_NONE_FONT * max(0, MAX_ITEMS_PER_ROW - len(line)))
					line.insert(0, SMALL_NONE_FONT * LEFT_PADDING)
					content.extend(deepcopy(line))
					content.extend(category_padding)
					for i in range(1, len(line)):
						selected = line[-i]
						if isinstance(selected, dict):
							selected["text"] = MEDIUM_NONE_FONT
					content.extend(["\n", *line, *category_padding, "\n"])

				# Add the 2 pixels border
				page_image = add_border(page_image, get_border_color(), BORDER_SIZE)

				# Save the image
				page_image.save(f"{SharedMemory.cache_path}/font/category/{file_name}.png")

			# Encode items
			elif isinstance(raw_data, Item):
				# Get all crafts
				crafts: list[JsonDict] = (raw_data.to_dict()["recipes"]).copy()
				crafts += generate_otherside_crafts(name, definitions_as_objects)
				crafts = remove_duplicate_furnace_crafts(crafts, name)
				crafts = remove_unknown_crafts(crafts)
				crafts = stp.unique_list(crafts)

				# Helper function to add count information to mining recipes
				def add_count_to_mining_recipe(mining_recipe: JsonDict, no_silk_drop_data: JsonDict | str) -> None:
					if isinstance(no_silk_drop_data, dict) and "count" in no_silk_drop_data:
						count_data: JsonDict | int = no_silk_drop_data["count"]
						if isinstance(count_data, dict):
							# Range of items like {"min": 2, "max": 8}
							if "min" in count_data and "max" in count_data:
								mining_recipe["result_count"] = f"{count_data['min']}-{count_data['max']}"
							elif "min" in count_data:
								mining_recipe["result_count"] = str(count_data["min"])
							elif "max" in count_data:
								mining_recipe["result_count"] = str(count_data["max"])
						else:
							# Single count value
							mining_recipe["result_count"] = str(count_data)

				# If there is NO_SILK_TOUCH_DROP, add a mining recipe for it
				no_silk_touch_drop: bool = bool(raw_data.get(NO_SILK_TOUCH_DROP, False))
				is_drop_of: list[str] = [
					i for i, d in Mem.definitions.items()
					if d.get(NO_SILK_TOUCH_DROP) and (
						(isinstance(d[NO_SILK_TOUCH_DROP], str) and d[NO_SILK_TOUCH_DROP] == name) or
						(isinstance(d[NO_SILK_TOUCH_DROP], dict) and cast(JsonDict, d[NO_SILK_TOUCH_DROP]).get("id") == name)
					)
				]

				# Add mining recipes for items that are drops from other ores
				content_added: bool = False
				if is_drop_of:
					for ore_name in is_drop_of:
						mining_recipe: JsonDict = {
							"type": "mining",
							"ingredient": Ingr(ore_name, Mem.ctx.project_id),  # The ore being mined
							"result": Ingr(name, Mem.ctx.project_id),  # This item is the result
						}
						add_count_to_mining_recipe(mining_recipe, Mem.definitions[ore_name][NO_SILK_TOUCH_DROP])
						crafts.insert(0, mining_recipe)

					# Generate the craft content
					content += generate_craft_content(crafts[0], name, page_font)
					content_added = True

				# Add mining recipe if this item has no_silk_touch_drop (it's an ore)
				if no_silk_touch_drop:
					no_silk_drop_data: JsonDict | str = raw_data[NO_SILK_TOUCH_DROP]
					result_format: str = no_silk_drop_data if isinstance(no_silk_drop_data, str) else no_silk_drop_data["id"]
					mining_recipe: JsonDict = {
						"type": "mining",
						"ingredient": Ingr(name, Mem.ctx.project_id),  # The ore being mined
						"result": Ingr(result_format, Mem.ctx.project_id),  # Proper ingredient format
					}
					add_count_to_mining_recipe(mining_recipe, no_silk_drop_data)

					# Insert at position 0 so it appears first and generate the craft content
					crafts.insert(0, mining_recipe)
					if not content_added:
						content += generate_craft_content(crafts[0], name, page_font)
						content_added = True

				# Else, if there are blue crafts, generate the content for the first craft
				elif not content_added:
					blue_crafts: list[JsonDict] = [craft for craft in crafts if not craft.get("result")]
					if blue_crafts:
						# Sort crafts by result_count in reverse order
						blue_crafts.sort(key=lambda craft: craft.get("result_count", 0), reverse=True)

						# Get the first craft and generate the content
						content += generate_craft_content(blue_crafts[0], name, page_font)
						content_added = True

					# Else, generate the content for the single item in a big box
					else:
						if page_font == "":
							page_font = get_page_font(number)
						generate_page_font(name, page_font, craft = None)
						component = get_item_component(name)
						component["text"] = NONE_FONT
						component["text"] *= 2
						content.append({"text": "", "font": FONT, "color": "white"})	# Make default font for every next component
						content.append({"text": titled, "font": "minecraft:default", "color": "black", "underlined": True})
						if not SharedMemory.use_dialog > 0:
							content.append(MEDIUM_NONE_FONT * 2 + page_font + "\n")
						else:
							content.append(page_font + "\n")

						for _ in range(4):
							if not SharedMemory.use_dialog > 0:
								content.append(MEDIUM_NONE_FONT * 2)
							content.append(component)
							content.append("\n")
						content_added = True

				## Add wiki information if any
				info_buttons: list[JsonDict] = []
				if name == "heavy_workbench":
					content.append([
						{"text":"\nEvery recipe that uses custom items ", "font":"minecraft:default", "color":"black"},
						{"text":"must", "color":"red", "underlined":True},
						{"text":" be crafted using the Heavy Workbench."}
					])
				else:
					if raw_data.get(WIKI_COMPONENT):
						found_event: JsonDict | None = None
						wiki_component: TextComponent | list[WikiButton] = raw_data[WIKI_COMPONENT]
						is_list_of_wiki_button: bool = isinstance(wiki_component, list) and any(isinstance(comp, WikiButton) for comp in wiki_component)
						wiki_buttons: list[TextComponent] = wiki_component if is_list_of_wiki_button else [wiki_component] # type: ignore

						# For each wiki button, add it to the info_buttons list
						for button in wiki_buttons:
							# Convert WikiButton to dict if needed
							if isinstance(button, WikiButton):
								button = button.to_dict()

							# Get click_event if any
							if isinstance(button, dict) and "click_event" in button:
								found_event = button["click_event"]
							elif isinstance(button, list):
								for comp in button:
									if isinstance(comp, dict) and "click_event" in comp:
										found_event = cast(JsonDict, comp["click_event"])
										break

							info_buttons.append({
								"text": WIKI_INFO_FONT + VERY_SMALL_NONE_FONT * 2,
								"hover_event": {
									"action": "show_text",
									"value": button
								}
							})
							if found_event:
								info_buttons[-1]["click_event"] = found_event

					# For each craft (except smelting dupes),
					previous_result: Any = None
					for i, craft in enumerate(crafts):
						if craft["type"] == CraftingShapelessRecipe.type:
							craft = convert_shapeless_to_shaped(craft)

						# Skip if result is same as previous result
						current_result: Any | None = craft.get("result")
						if current_result and current_result == previous_result and craft["type"] != "mining":
							continue
						previous_result = current_result

						# Get breaklines
						breaklines = 3
						if "shape" in craft:
							breaklines = max(2, max(len(craft["shape"]), len(craft["shape"][0])))

						if not SharedMemory.high_resolution:
							craft_font = get_next_font()	# Unique used font for the craft
							generate_page_font(name, craft_font, craft, output_name = f"{name}_{i+1}")
							hover_text: list[TextComponent] = [{"text":""}]
							hover_text.append({"text": craft_font + "\n\n" * breaklines, "font": FONT, "color": "white"})
						else:
							craft_content: list[TextComponent] = generate_craft_content(craft, name, "", in_lore=True)
							craft_content = [craft_content[0], *craft_content[2:]]	# Remove craft title
							remove_events(craft_content)
							for k, v in HOVER_EQUIVALENTS.items():
								if isinstance(craft_content[1], str):
									craft_content[1] = craft_content[1].replace(k, v)
							hover_text = [{"text":""}, craft_content]

						# Add recipe type title
						if craft["type"] not in CRAFTING_RECIPES_TYPES:
							recipe_type_names: dict[str, str] = {
								"mining": "Mining",
								CraftingShapedRecipe.type: "Shaped Recipe",
								CraftingShapelessRecipe.type: "Shapeless Recipe",
								SmeltingRecipe.type: "Smelting",
								BlastingRecipe.type: "Blasting",
								SmokingRecipe.type: "Smoking",
								CampfireCookingRecipe.type: "Campfire Cooking",
								StonecuttingRecipe.type: "Stonecutting",
								SmithingTransformRecipe.type: "Smithing Transform",
								SmithingTrimRecipe.type: "Smithing Trim",
								PulverizingRecipe.type: "(SimplEnergy) Pulverizing",
								AwakenedForgeRecipe.type: "(Stardust Fragment) Awakened Forge",
							}
							recipe_title = recipe_type_names.get(craft["type"], craft["type"].replace("_", " ").title())
							hover_text.append({"text": f"\n{recipe_title}", "color": "yellow"})

						# Append ingredients
						if craft["type"] == "mining":
							# For mining recipes, show what is being mined and what it drops
							ore_name = Ingr(craft["ingredient"]).to_name()
							result_name = Ingr(craft["result"]).to_name()
							hover_text.append({"text": "\n- Mine: ", "color": "gray"})
							hover_text.append({"text": ore_name, "color": "gray"})
							result_count = craft.get("result_count", "1")
							hover_text.append({"text": f"\n- Drops: x{result_count} ", "color": "gray"})
							hover_text.append({"text": result_name, "color": "gray"})
						elif craft.get("ingredient"):
							id = Ingr(craft["ingredient"]).to_name()
							hover_text.append({"text": "\n- x1 ", "color": "gray"})
							hover_text.append({"text": id, "color": "gray"})
						elif craft.get("ingredients"):

							# If it's a shaped crafting
							if isinstance(craft["ingredients"], dict):
								for k, v in craft["ingredients"].items():
									id = Ingr(v).to_name()
									count = sum([line.count(k) for line in craft["shape"]])
									hover_text.append({"text": f"\n- x{count} ", "color": "gray"})
									hover_text.append({"text": id, "color": "gray"})

							# If it's shapeless
							elif isinstance(craft["ingredients"], list):
								ids: dict[str, int] = {}	# {id: count}
								for ingr in craft["ingredients"]:
									id = Ingr(ingr).to_name()
									if id not in ids:
										ids[id] = 0
									ids[id] += ingr.get("count", 1)
								for id, count in ids.items():
									hover_text.append({"text": f"\n- x{count} ", "color": "gray"})
									hover_text.append({"text": id, "color": "gray"})
						else:
							# Smithing crafts
							if craft.get("base"):
								id = Ingr(craft["base"]).to_name()
								hover_text.append({"text": "\n- Base: ", "color": "gray"})
								hover_text.append({"text": id, "color": "gray"})
							if craft.get("template"):
								id = Ingr(craft["template"]).to_name()
								hover_text.append({"text": "\n- Template: ", "color": "gray"})
								hover_text.append({"text": id, "color": "gray"})
							if craft.get("addition"):
								id = Ingr(craft["addition"]).to_name()
								hover_text.append({"text": "\n- Addition: ", "color": "gray"})
								hover_text.append({"text": id, "color": "gray"})
							if craft.get("pattern"):
								pattern_name = str(craft["pattern"]).replace("minecraft:", "").replace("_", " ").title()
								hover_text.append({"text": "\n- Pattern: ", "color": "gray"})
								hover_text.append({"text": pattern_name, "color": "gray"})

						# Add the craft to the content
						result_or_ingredient = WIKI_RESULT_OF_CRAFT_FONT if "result" not in craft else generate_wiki_font_for_ingr(name, craft)
						info_buttons.append({
							"text": result_or_ingredient + VERY_SMALL_NONE_FONT * 2,
							"hover_event": {
								"action": "show_text",
								"value": hover_text
							},
							"priority": craft.get("manual_priority", 1)
						})

						# If there is a result to the craft, try to add the click_event that change to that page
						craft_result: str = "" if "result" not in craft else Ingr(craft["result"]).to_id(add_namespace=False)
						if craft_result and craft_result != name:
							if craft_result in Mem.definitions:
								page_number: int = get_page_number(craft_result)
								if page_number != -1:
									info_buttons[-1]["click_event"] = {
										"action": "change_page",
										"page": page_number,
									}

						# Else, try to add the click_event that change to that page
						else:
							# If there is only one ingredient, link to it
							craft_ingredient: str = ""
							if craft.get("ingredient"):
								craft_ingredient = Ingr(craft["ingredient"]).to_id(add_namespace=False)
							elif craft.get("ingredients") and isinstance(craft["ingredients"], list) and len(craft["ingredients"]) == 1:
								craft_ingredient = Ingr(craft["ingredients"][0]).to_id(add_namespace=False)
							elif craft.get("ingredients") and isinstance(craft["ingredients"], dict) and len(craft["ingredients"]) == 1:
								craft_ingredient = Ingr(next(iter(craft["ingredients"].values()))).to_id(add_namespace=False)
							if craft_ingredient and craft_ingredient in Mem.definitions and craft_ingredient != name:
								page_number: int = get_page_number(craft_ingredient)
								if page_number != -1:
									info_buttons[-1]["click_event"] = {
										"action": "change_page",
										"page": page_number,
										"blue_craft": craft_result == "" # No result = blue craft
									}

				# Add wiki buttons 5 by 5
				if info_buttons:

					# Retrieve buttons limit depending on number of lines in the content (more lines = less buttons)
					buttons_limit: int = 20 if str(content).count("\\n") <= 6 else 15
					if SharedMemory.use_dialog > 0:
						buttons_limit = 42	# Higher limit

					# If too many buttons, remove all the blue ones (blue_craft) except the last one
					if len(info_buttons) > buttons_limit:
						first_index: int = 0 if not raw_data.get(WIKI_COMPONENT) else 1
						last_index: int = -1
						for i, button in enumerate(info_buttons):
							if button.get("click_event", {}).get("blue_craft", False) and i != first_index:
								last_index = i

						# If there are more than 1 blue button, remove them except the last one
						if (last_index - first_index) > 1:
							info_buttons = info_buttons[:first_index] + info_buttons[last_index:]

						# If there are more than buttons_limit buttons, remove lowest priority ones until there is buttons_limit left
						while len(info_buttons) > buttons_limit:
							lowest_priority: JsonDict = min(info_buttons[::-1], key=lambda x: x.get("priority", 1))
							info_buttons.remove(lowest_priority)

						# Keep only the last buttons_limit buttons (maximum that can be displayed with 5 per line and 4 lines)
						info_buttons = info_buttons[:buttons_limit]

					for button in info_buttons:
						click_event: JsonDict = button.get("click_event", {})
						if click_event.get("blue_craft", False):
							del click_event["blue_craft"]
						if button.get("priority"):
							del button["priority"]

					# Add a breakline only if there aren't too many breaklines already
					content.append("\n")

					last_i = 0
					buttons_per_line = 5 if not SharedMemory.use_dialog > 0 else 6
					for i, button in enumerate(info_buttons):
						last_i = i
						# Duplicate line and add breakline
						if i % buttons_per_line == 0 and i != 0:
							# Remove VERY_SMALL_NONE_FONT from last button to prevent automatic break line
							last_content = cast(JsonDict, content[-1])
							last_content["text"] = last_content["text"].replace(VERY_SMALL_NONE_FONT, "")

							# Re-add last buttons (for good hover_event) but we replace the wiki font by the small font
							content += ["\n"] + [cast(JsonDict, x).copy() for x in content[-buttons_per_line:]]
							for j in range(buttons_per_line):
								selected_content = cast(JsonDict, content[-buttons_per_line + j])
								selected_content["text"] = WIKI_NONE_FONT + VERY_SMALL_NONE_FONT * (2 if j != (buttons_per_line - 1) else 0)
							content.append("\n")
						content.append(button)

					# Duplicate the last line if not done yet
					if last_i % buttons_per_line != 0 or last_i == 0:
						last_i = last_i % buttons_per_line + 1

						# Remove VERY_SMALL_NONE_FONT from last button to prevent automatic break line
						last_content = cast(JsonDict, content[-1])
						last_content["text"] = last_content["text"].replace(VERY_SMALL_NONE_FONT, "")

						content += ["\n"] + [cast(JsonDict, x).copy() for x in content[-last_i:]]
						for j in range(last_i):
							selected_content = cast(JsonDict, content[-last_i + j])
							selected_content["text"] = WIKI_NONE_FONT + VERY_SMALL_NONE_FONT * (2 if j != (last_i - 1) else 0)

			# Add page to the book
			book_content.append(content)
			pass

		# TODO: Make it compatible with multithreading (ordering problem only I think)
		#stp.multithreading(encode_page, SharedMemory.manual_pages, desc="Creating manual pages", max_workers=-1.0)
		for page in stp.colored_for_loop(SharedMemory.manual_pages, desc="Creating manual pages"):
			encode_page(page)

		## Add categories page
		content: list[TextComponent] = []
		file_name = "categories_page"
		page_font = get_page_font(1)
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/category/{file_name}.png", "ascent": 1, "height": 131, "chars": [page_font]})
		content.append({"text": "", "font": FONT, "color": "white"})	# Make default font for every next component
		content.append({"text": "➤ ", "font": "minecraft:default", "color": "black"})
		content.append({"text": "Category browser\n", "font": "minecraft:default", "color": "black", "underlined": True})
		content.append(SMALL_NONE_FONT * LEFT_PADDING + page_font + "\n")

		# Prepare image and line list
		page_image = Image.new("RGBA", (256, 256), (0, 0, 0, 0))
		x, y = 2, 2	# Prevision for global border and implicit border
		line: list[TextComponent] = []

		# For each item in the category, get its page number and texture, then add it to the image
		max_items_reached: bool = False
		for page_content in SharedMemory.manual_pages:
			if page_content["type"] == CATEGORY:
				item = page_content["raw_data"][0]

				# Get item texture
				texture_path = f"{SharedMemory.cache_path}/items/{Mem.ctx.project_id}/{item}.png"
				if os.path.exists(texture_path):
					item_image = Image.open(texture_path)
				else:
					stp.warning(f"Missing texture at '{texture_path}', using empty texture")
					item_image = Image.new("RGBA", (1, 1), (0, 0, 0, 0))
				if not SharedMemory.high_resolution:
					resized = careful_resize(item_image, 32)
					high_res_font = None
				else:
					resized = Image.new("RGBA", (1, 1), (0, 0, 0, 0))	# Empty texture to use for category page
					high_res_font = generate_high_res_font(item, item_image)

				# Paste the simple case and the item_image
				page_image.paste(simple_case, (x, y))
				mask = resized.convert("RGBA").split()[3]
				page_image.paste(resized, (x + 2, y + 2), mask)
				x += simple_case.size[0]

				# Add the click_event part to the line and add the 2 times the line if enough items
				component = get_item_component(item, ["item_name"])
				component["hover_event"]["components"]["item_name"] = {"text": page_content["name"], "color": "white"}
				component["click_event"]["page"] = page_content["number"]
				if not SharedMemory.high_resolution:
					component["text"] = MEDIUM_NONE_FONT
				else:
					component["text"] = high_res_font
				line.append(component)
				if len(line) == MAX_ITEMS_PER_ROW:
					max_items_reached = True
					line.insert(0, SMALL_NONE_FONT * LEFT_PADDING)
					content += [*deepcopy(line), *category_padding, "\n"]
					for i in range(1, len(line)):
						selected = cast(JsonDict, line[-i])
						selected["text"] = MEDIUM_NONE_FONT
					content += [*line, *category_padding, "\n"]
					line = []
					x = 2
					y += simple_case.size[1]

		# If remaining items in the line, add them
		if len(line) > 0:
			if SharedMemory.use_dialog > 0 and max_items_reached:
				line.append(MEDIUM_NONE_FONT * max(0, MAX_ITEMS_PER_ROW - len(line)))
			line.insert(0, SMALL_NONE_FONT * LEFT_PADDING)
			content += [*deepcopy(line), *category_padding, "\n"]
			for i in range(1, len(line)):
				selected = cast(JsonDict | str, line[-i])
				if isinstance(selected, dict):
					selected["text"] = MEDIUM_NONE_FONT
			content += [*line, *category_padding, "\n"]

		# Add the 2 pixels border
		page_image = add_border(page_image, get_border_color(), BORDER_SIZE)

		# Save the image and add the page to the book
		page_image.save(f"{SharedMemory.cache_path}/font/category/{file_name}.png")
		book_content.insert(0, content)


		## Append introduction page
		intro_content: list[TextComponent] = [{"text":""}]
		page_font = get_page_font(0)
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/page/_logo.png", "ascent": 0, "height": 40, "chars": [page_font]})
		intro_content.append({"text": manual_name + "\n", "underlined": True})
		intro_content.append({"text": MEDIUM_NONE_FONT * (0 if SharedMemory.use_dialog > 0 else 2) + page_font, "font": FONT, "color": "white"})

		# Create the image and load Minecraft font

		icon_path = find_pack_png()
		assert icon_path and os.path.exists(icon_path), "Missing pack.png in your working tree (needed for the manual)"
		logo = Image.open(icon_path)
		logo = careful_resize(logo, 256)

		# Write the introduction text
		intro_content.append({"text": "\n" * 6})
		first_page_config: TextComponent = manual_config.get('first_page_text', "")
		intro_content.append([{"text": "", "font": "minecraft:default", "color": "black"}, first_page_config])

		# Save image and insert in the manual pages
		logo.save(f"{SharedMemory.cache_path}/font/page/_logo.png")
		book_content.insert(0, intro_content)

		## Optimize the book size
		book_content_deepcopy: list[TextComponent] = deepcopy(book_content)	# Deepcopy to avoid sharing same components (such as click_event)
		book_content = cast(list[list[TextComponent]], list(optimize_element(book_content_deepcopy)))

		## Insert at 2nd page the awakened forge
		if has_forge_3x3 or has_forge_3x4:
			book_content.insert(1, get_stardust_forge_page())

			# Increase every change_page click event by 1
			for page_content in book_content:
				for component in page_content:
					if isinstance(component, dict):
						if "click_event" in component and component["click_event"].get("action") == "change_page":
							current_value: int = int(component["click_event"]["page"])
							component["click_event"]["page"] = current_value + 1
			for p in SharedMemory.manual_pages:
				p["number"] += 1

			# Insert heavy workbench page number in the manual pages
			SharedMemory.manual_pages.insert(2, {"number": 2, "name": "awakened_forge", "raw_data": {}, "type": "item"})

		## Insert at 2nd page the heavy workbench
		if "heavy_workbench" in Mem.definitions:
			heavy_workbench_page = book_content.pop(-1)
			book_content.insert(1, heavy_workbench_page)

			# Increase every change_page click event by 1
			for page_content in book_content:
				for component in page_content:
					if isinstance(component, dict):
						if "click_event" in component and component["click_event"].get("action") == "change_page":
							current_value: int = int(component["click_event"]["page"])
							component["click_event"]["page"] = current_value + 1
			for p in SharedMemory.manual_pages:
				p["number"] += 1

			# Insert heavy workbench page number in the manual pages
			SharedMemory.manual_pages.insert(2, {"number": 2, "name": "heavy_workbench", "raw_data": {}, "type": "item"})

		# Add fonts
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/none.png", "ascent": 8, "height": 20, "chars": [NONE_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/none.png", "ascent": 8, "height": 18, "chars": [MEDIUM_NONE_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/none.png", "ascent": 7, "height": 7, "chars": [SMALL_NONE_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/none.png", "ascent": 0, "height": 2, "chars": [VERY_SMALL_NONE_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/none.png", "ascent": 0, "height": 1, "chars": [MICRO_NONE_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/none.png", "ascent": 7, "height": 16, "chars": [WIKI_NONE_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/invisible_item.png", "ascent": 7, "height": 16, "chars": [INVISIBLE_ITEM_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/wiki_information.png", "ascent": 8, "height": 16, "chars": [WIKI_INFO_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/wiki_result_of_craft.png", "ascent": 8, "height": 16, "chars": [WIKI_RESULT_OF_CRAFT_FONT]})
		SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/wiki_ingredient_of_craft.png", "ascent": 8, "height": 16, "chars": [WIKI_INGR_OF_CRAFT_FONT]})
		if SharedMemory.high_resolution:
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/shaped_3x3.png", "ascent": 1, "height": 58, "chars": [SHAPED_3X3_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/shaped_2x2.png", "ascent": 1, "height": 58, "chars": [SHAPED_2X2_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/furnace.png", "ascent": 1, "height": 58, "chars": [FURNACE_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/stonecutting.png", "ascent": 4, "height": 58, "chars": [STONECUTTING_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/pulverizing.png", "ascent": 4, "height": 58, "chars": [PULVERIZING_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/mining.png", "ascent": 4, "height": 58, "chars": [MINING_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/shaped_3x3.png", "ascent": -4, "height": 58, "chars": [HOVER_SHAPED_3X3_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/shaped_2x2.png", "ascent": -2, "height": 58, "chars": [HOVER_SHAPED_2X2_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/furnace.png", "ascent": -3, "height": 58, "chars": [HOVER_FURNACE_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/stonecutting.png", "ascent": -3, "height": 58, "chars": [HOVER_STONECUTTING_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/pulverizing.png", "ascent": -3, "height": 58, "chars": [HOVER_PULVERIZING_FONT]})
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/mining.png", "ascent": -3, "height": 58, "chars": [HOVER_MINING_FONT]})
			if has_forge_3x3:
				SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/awakened_forge_3x3.png", "ascent": 9, "height": 74, "chars": [AWAKENED_3X3_FONT]})
				SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/awakened_forge_3x3.png", "ascent": 4, "height": 74, "chars": [HOVER_AWAKENED_3X3_FONT]})
			if has_forge_3x4:
				SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/awakened_forge_3x4.png", "ascent": 9, "height": 74, "chars": [AWAKENED_3X4_FONT]})
				SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/awakened_forge_3x4.png", "ascent": 4, "height": 74, "chars": [HOVER_AWAKENED_3X4_FONT]})
			if has_forge_3x3 or has_forge_3x4:
				SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/awakened_forge_1.png", "ascent": 8, "height": 56, "chars": [AWAKENED_FORGE_STRUCT_FONT[0]]})
				SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/awakened_forge_2.png", "ascent": 8, "height": 56, "chars": [AWAKENED_FORGE_STRUCT_FONT[1]]})
		if SharedMemory.use_dialog > 0:
			SharedMemory.font_providers.append({"type":"bitmap","file":f"{Mem.ctx.project_id}:font/book.png", "ascent": 25, "height": 300, "chars": [BOOK_FONT]})
		fonts = {"providers": SharedMemory.font_providers}
		with stp.super_open(f"{SharedMemory.cache_path}/font/manual.json", "w") as f:
			f.write(stp.json_dump(fonts))

		# Debug book_content
		json_dump_path: str = manual_config.get("json_dump_path", "")
		if json_dump_path:
			with stp.super_open(json_dump_path, "w") as f:
				f.write(stp.json_dump(book_content))
			stp.debug(f"Debug book_content at '{stp.relative_path(json_dump_path)}'")

		# Generate showcase images if requested
		showcase_image: int = manual_config.get("showcase_image", 3)
		if showcase_image > 0:
			generate_showcase_images(showcase_image, categories, simple_case)


	# Copy the font provider and the generated textures to the resource pack
	Mem.ctx.assets[Mem.ctx.project_id].fonts["manual"] = Font(source_path=f"{SharedMemory.cache_path}/font/manual.json")
	for folder in ["category", "page", "wiki_icons", *(["high_res"] if SharedMemory.high_resolution else [])]:
		for file in os.listdir(f"{SharedMemory.cache_path}/font/{folder}"):
			file_path: str = f"{SharedMemory.cache_path}/font/{folder}/{file}"
			no_extension: str = os.path.splitext(file)[0]
			if file.endswith(".png") and os.path.isfile(file_path):
				Mem.ctx.assets[Mem.ctx.project_id].textures[f"font/{folder}/{no_extension}"] = Texture(source_path=file_path)

	# Verify font providers and textures
	for fp in SharedMemory.font_providers:
		if "file" in fp:
			path: str = fp["file"]
			path = os.path.splitext(path.split(":", 1)[-1])[0]  # Remove namespace and extension
			if not Mem.ctx.assets[Mem.ctx.project_id].textures.get(path):
				stp.error(f"Missing font provider at '{path}' for {fp})")
			if len(fp["chars"]) < 1 or (len(fp["chars"]) == 1 and not fp["chars"][0]):
				stp.error(f"Font provider '{path}' has no chars")

	# Generate dialog files if needed
	if SharedMemory.use_dialog > 0:
		generate_dialogs(book_content)

	# Finally, prepend the manual to the definitions
	if SharedMemory.use_dialog != 2:
		manual_already_exists: bool = "manual" in Mem.definitions
		manual_obj = Item(
			id="manual",
			base_item="minecraft:written_book",
			components={
				"written_book_content": {
					"title": manual_name,
					"author": Mem.ctx.project_author,
					"pages": book_content,
				},
				"item_model": f"{Mem.ctx.project_id}:manual",
				"item_name": manual_name,
				"enchantment_glint_override": False,
				"max_stack_size": 16
			}
		)
		if SharedMemory.use_dialog > 0:
			del manual_obj.components["written_book_content"]
			manual_obj.components["lore"] = [{"text": f"by {Mem.ctx.project_author}", "color": "gray", "italic": False}]
		if manual_already_exists:
			current_def: Item = Item.from_id("manual")
			current_def.components = super_merge_dict(manual_obj.components, current_def.components)
			current_def.base_item = manual_obj.base_item
			Mem.definitions["manual"] = current_def
		add_item_name_and_lore_if_missing(black_list=[item for item in Mem.definitions if item != "manual"])
		add_private_custom_data_for_namespace(black_list=[item for item in Mem.definitions if item != "manual"])

		# Add the model to the resource pack if it doesn't already exist
		if not manual_already_exists:
			textures_folder: str = stp.relative_path(Mem.ctx.meta.get("stewbeet", {}).get("textures_folder", ""))
			textures: dict[str, str] = {
				stp.clean_path(str(p)).split("/")[-1]: stp.relative_path(str(p))
				for p in Path(textures_folder).rglob("*.png")
			}
			AutoModel.from_definitions(manual_obj, textures).process()

		# Repair the recipes for the manual
		VanillaRecipeHandler().generate_recipes(override=["manual"])

	# Remove the heavy workbench from the definitions
	if OFFICIAL_LIBS["smithed.crafter"]["is_used"]:
		del Mem.definitions["heavy_workbench"]
		del Mem.ctx.assets[Mem.ctx.project_id].textures["item/heavy_workbench"]
		del Mem.ctx.assets[Mem.ctx.project_id].models["item/heavy_workbench"]
		del Mem.ctx.assets[Mem.ctx.project_id].item_models["heavy_workbench"]

	# Register of the manual in the universal manual
	first_page: str = json.dumps(book_content[0], ensure_ascii=False)
	for r in [("\\n", "\\\\n"), (', "underlined": true','')]:
		first_page = first_page.replace(*r)
	write_load_file(
f"""
# Register the manual to the universal manual
execute unless data storage stewbeet:main universal_manual run data modify storage stewbeet:main universal_manual set value []
data remove storage stewbeet:main universal_manual[{{"name":"{Mem.ctx.project_name}"}}]
data modify storage stewbeet:main universal_manual append value {{"name":"{Mem.ctx.project_name}","loot_table":"{Mem.ctx.project_id}:i/manual","hover":{first_page}}}
""")
	pass




"""
Macro argument parsing utilities for Minecraft function headers.

This module handles extracting macro variables from function content
by finding lines that start with '$' and extracting $(variable_name) patterns.
"""

# Imports
import re


# Functions
def extract_macro_variables(content: str) -> list[str]:
    """ Extract all macro variables from function content.

    Args:
        content (str): The function content

    Returns:
        list[str]: List of unique macro variable names found in the content, in order of first appearance

    Examples:
        Basic example:
        >>> content = '''
        ... $execute if score @s temp matches $(slot) run say Found
        ... $data modify entity @s Items[{Slot:$(slot)b}].id set value "$(id)"
        ... $give @p $(id) $(count)
        ... '''
        >>> extract_macro_variables(content)
        ['slot', 'id', 'count']

        From StardustFragment teleport_to:
        >>> content = '$execute in $(dimension) run tp @s $(x) $(y).6 $(z) $(yaw) $(pitch)'
        >>> extract_macro_variables(content)
        ['dimension', 'x', 'y', 'z', 'yaw', 'pitch']

        From SimplEnergy get_max_stack_size:
        >>> content = '''
        ... $execute if items entity @s container.$(result) *[minecraft:max_stack_size=64] run return 64
        ... $execute if items entity @s container.$(result) *[minecraft:max_stack_size=16] run return 16
        ... '''
        >>> extract_macro_variables(content)
        ['result']

        From SimplEnergy update_energy_lore with multiple vars:
        >>> content = '$data modify storage energy:temp list[0] value [$(part_1),$(part_2),$(scale)]'
        >>> extract_macro_variables(content)
        ['part_1', 'part_2', 'scale']
    """
    # Preserve first-occurrence order of macro variables as they appear in the
    # function content. Using a list and membership checks keeps the order
    # stable and mirrors what the user sees in the source file.
    macro_vars: list[str] = []

    # Split content into lines
    lines: list[str] = content.split("\n")

    for line in lines:
        # Check if line starts with '$' (after stripping whitespace and comments)
        stripped: str = line.strip()
        if stripped.startswith("$"):
            # Find all $(variable_name) patterns in the line
            matches: list[str] = re.findall(r'\$\((\w+)\)', line)
            for m in matches:
                if m not in macro_vars:
                    macro_vars.append(m)

    return macro_vars


def is_macro_function(content: str) -> bool:
    """ Check if a function is a macro function.

    A function is considered a macro function if it contains at least one line
    that starts with '$'.

    Args:
        content (str): The function content

    Returns:
        bool: True if the function is a macro function, False otherwise

    Examples:
        Non-macro function:
        >>> is_macro_function("say Hello")
        False

        Simple macro:
        >>> is_macro_function("$say $(message)")
        True

        Macro with comment:
        >>> is_macro_function("# Comment\\n$execute run say $(text)")
        True

        From StardustFragment:
        >>> is_macro_function("$execute in $(dimension) run tp @s $(x) $(y) $(z)")
        True

        From SimplEnergy:
        >>> is_macro_function("$execute if items entity @s container.$(result) *[max_stack_size=64] run return 64")
        True
    """
    lines: list[str] = content.split("\n")
    for line in lines:
        stripped: str = line.strip()
        if stripped.startswith("$"):
            return True
    return False




# Imports
from __future__ import annotations

import stouputils as stp
from beet.core.utils import JsonDict


# Utility function to convert result_count to string suffix
@stp.simple_cache
def result_count_to_suffix(result_count: int | JsonDict) -> str:
	""" Convert a result count to a string suffix for loot table paths

	Args:
		result_count (int | JsonDict): The count of the result item, can be an int or a dict for random counts
			ex: 1
			ex: {"type": "minecraft:uniform","min": 4,"max": 6}
	Returns:
		str: The suffix string, ex: "" or "_x5" or "_x4to6"
	"""
	if isinstance(result_count, int):
		if result_count > 1:
			return f"_x{result_count}"
		return ""
	elif hasattr(result_count, "get"):
		minimum = result_count.get("min", 1)
		maximum = result_count.get("max", 1)
		if maximum > 1:
			return f"_x{minimum}to{maximum}"
		elif minimum > 1:
			return f"_x{minimum}"
	return ""





# Imports
from beet import Context
from beet.core.utils import JsonDict

from .quick_sort import generate_quick_sort
from .selection_sort import generate_selection_sort


# Functions
def generate_sorter(ctx: Context, sorter: JsonDict):
	""" Generate sorting functions based on the specified algorithm.

	Currently an abstraction layer in case more algorithms are added in the future.
	Routes to the appropriate algorithm-specific generator function.

	Args:
		ctx (Context): The beet context for generating functions.
		sorter (dict): Configuration dictionary containing sorting parameters.
	"""
	# Assertions to validate sorter configuration
	assert isinstance(sorter.get("functions_location"), str), "'functions_location' must be a string, e.g. 'switch:stats/minigame/sort_leaderboard'."
	assert isinstance(sorter.get("key"), str), "'key' must be a string, e.g. 'count' if you list is looking like [{'id':'...', count:54}, ...]."
	assert isinstance(sorter.get("to_sort"), dict), "'to_sort' must be a dictionary."
	assert isinstance(sorter["to_sort"].get("storage"), str), "'to_sort.storage' must be a string, e.g. 'switch:stats'"
	assert isinstance(sorter["to_sort"].get("target"), str), "'to_sort.target' must be a string, e.g. 'all.modes.sheepwars.played'."

	# Switch case
	algorithm: str = sorter["algorithm"]
	match(algorithm):
		case "selection_sort":
			generate_selection_sort(ctx, sorter)
		case "quick_sort":
			generate_quick_sort(ctx, sorter)
		case _:
			raise ValueError(f"Unknown sorting algorithm: {algorithm}")





# Imports
import os
import time
from pathlib import Path
from zipfile import ZIP_DEFLATED, ZipFile, ZipInfo

import stouputils as stp
from beet import Context

from ...core.constants import OFFICIAL_LIBS
from ...dependencies import OFFICIAL_LIBS_PATH
from ..archive import get_consistent_timestamp
from ..initialize.source_lore_font import find_pack_png


# Weld datapack
@stp.handle_error
def weld_datapack(ctx: Context, dest_path: str) -> float:
	""" Merge the datapack and libs into one file using Weld
	Args:
		ctx (Context): The beet context
		dest_path (str): The path to the destination file
	Returns:
		float: The time it took to merge the datapack and libs
	"""
	start_time: float = time.perf_counter()

	# Get configuration from context
	stewbeet_config = ctx.meta.get("stewbeet", {})
	libs_folder = stewbeet_config.get("libs_folder", "libs")
	project_name_simple = ctx.project_name.replace(" ", "")

	# Get all paths to merge
	datapacks_to_merge = [
		str(Path(str(ctx.output_directory)) / f"{project_name_simple}_datapack.zip")
	]
	if libs_folder and os.path.exists(libs_folder):
		datapacks_to_merge.append(f"{libs_folder}/datapack/*.zip")

	# Add the used official libs
	for lib in OFFICIAL_LIBS.values():
		if lib["is_used"]:
			name: str = lib["name"]
			path: str = f"{OFFICIAL_LIBS_PATH}/datapack/{name}.zip"
			if os.path.exists(path):
				datapacks_to_merge.append(path)

	# Skip welding if there are less than 2 datapacks to merge
	if len(datapacks_to_merge) < 2:
		stp.warning(f"No datapacks or libs to merge for {dest_path}. Skipping weld.")
		return time.perf_counter() - start_time

	# Weld all datapacks
	output_dir = os.path.dirname(dest_path)
	output = os.path.basename(dest_path.replace(".zip", "_temporary.zip"))
	from smithed.weld.toolchain.cli import weld  # pyright: ignore[reportMissingTypeStubs]
	stp.silent(weld)(datapacks_to_merge, Path(output_dir), Path(output), log = "error")

	# Get the consistent timestamp
	constant_time = get_consistent_timestamp(ctx)

	# Make the new zip file with fixed pack.mcmeta and pack.png
	with ZipFile(dest_path.replace(".zip", "_temporary.zip"), "r") as temp_zip:

		# Open the final destination zip file for writing
		with ZipFile(dest_path, "w", compression=ZIP_DEFLATED, compresslevel=6) as zip:

			# Iterate through all files in the temporary zip, and exclude pack.mcmeta and pack.png
			for file in temp_zip.namelist():
				if file not in ["pack.mcmeta", "pack.png"]:
					info: ZipInfo = ZipInfo(file)
					info.compress_type = ZIP_DEFLATED
					info.date_time = constant_time
					zip.writestr(info, temp_zip.read(file))

			# Add the fixed pack.mcmeta to the final zip with constant_time
			if hasattr(ctx.data, 'mcmeta') and ctx.data.mcmeta:
				info: ZipInfo = ZipInfo("pack.mcmeta")
				info.compress_type = ZIP_DEFLATED
				info.date_time = constant_time
				zip.writestr(info, ctx.data.mcmeta.text.encode('utf-8'))

			# Check if pack.png exists and add it to the final zip if it does
			pack_png_path = find_pack_png()
			if pack_png_path:
				info: ZipInfo = ZipInfo("pack.png")
				info.compress_type = ZIP_DEFLATED
				info.date_time = constant_time
				with open(pack_png_path, "rb") as f:
					zip.writestr(info, f.read())

	# Remove temp file
	os.remove(dest_path.replace(".zip","_temporary.zip"))

	# Return the time it took to merge the datapack and libs
	return time.perf_counter() - start_time


# Weld resource pack
@stp.handle_error
def weld_resource_pack(ctx: Context, dest_path: str) -> float:
	""" Merge the resource pack and libs into one file using Weld
	Args:
		ctx (Context): The beet context
		dest_path (str): The path to the destination file
	Returns:
		float: The time it took to merge the resource pack and libs
	"""
	start_time: float = time.perf_counter()

	# Get configuration from context
	stewbeet_config = ctx.meta.get("stewbeet", {})
	libs_folder = stewbeet_config.get("libs_folder", "libs")
	project_name_simple = ctx.project_name.replace(" ", "")

	# Get all paths to merge
	resource_packs_to_merge = [
		str(Path(str(ctx.output_directory)) / f"{project_name_simple}_resource_pack.zip")
	]
	if libs_folder and os.path.exists(libs_folder):
		resource_packs_to_merge.append(f"{libs_folder}/resource_pack/*.zip")

	# Add the used official libs
	for lib in OFFICIAL_LIBS.values():
		if lib["is_used"]:
			name: str = lib["name"]
			path: str = f"{OFFICIAL_LIBS_PATH}/resource_pack/{name}.zip"
			if os.path.exists(path):
				resource_packs_to_merge.append(path)

	# Skip welding if there are less than 2 resource packs to merge
	if len(resource_packs_to_merge) < 2:
		stp.warning(f"No resource packs or libs to merge for {dest_path}. Skipping weld.")
		return time.perf_counter() - start_time

	# Weld all resource packs
	output_dir = os.path.dirname(dest_path)
	output = os.path.basename(dest_path.replace(".zip", "_temporary.zip"))
	from smithed.weld.toolchain.cli import weld  # pyright: ignore[reportMissingTypeStubs]
	stp.silent(weld)(resource_packs_to_merge, Path(output_dir), Path(output), log = "error")

	# Get the consistent timestamp
	constant_time = get_consistent_timestamp(ctx)

	# Make the new zip file with fixed pack.mcmeta and pack.png
	with ZipFile(dest_path.replace(".zip", "_temporary.zip"), "r") as temp_zip:
		# Open the final destination zip file for writing
		with ZipFile(dest_path, "w", compression=ZIP_DEFLATED, compresslevel=6) as zip:
			# Iterate through all files in the temporary zip, and exclude pack.mcmeta and pack.png
			for file in temp_zip.namelist():
				if file not in ["pack.mcmeta", "pack.png"]:
					info: ZipInfo = ZipInfo(file)
					info.compress_type = ZIP_DEFLATED
					info.date_time = constant_time
					zip.writestr(info, temp_zip.read(file))

			# Add the fixed pack.mcmeta to the final zip with constant_time
			if hasattr(ctx.assets, 'mcmeta') and ctx.assets.mcmeta:
				info: ZipInfo = ZipInfo("pack.mcmeta")
				info.compress_type = ZIP_DEFLATED
				info.date_time = constant_time
				zip.writestr(info, ctx.assets.mcmeta.text.encode('utf-8'))

			# Check if pack.png exists and add it to the final zip if it does
			pack_png_path = find_pack_png()
			if pack_png_path:
				info: ZipInfo = ZipInfo("pack.png")
				info.compress_type = ZIP_DEFLATED
				info.date_time = constant_time
				with open(pack_png_path, "rb") as f:
					zip.writestr(info, f.read())

	# Remove temp file
	os.remove(dest_path.replace(".zip","_temporary.zip"))

	# Return the time it took to merge the resource pack and libs
	return time.perf_counter() - start_time





# Imports
from __future__ import annotations

from typing import Any, cast

import stouputils as stp
from beet import LootTable
from beet.core.utils import JsonDict

from ..__memory__ import Mem
from ..utils.io import set_json_encoder
from ..utils.loot_table import result_count_to_suffix
from ..utils.text_component import item_id_to_name

# Recipes constants
FURNACES_RECIPES_TYPES: tuple[str, ...] = ("smelting", "blasting", "smoking", "campfire_cooking")
CRAFTING_RECIPES_TYPES: tuple[str, ...] = ("crafting_shaped", "crafting_shapeless")
OTHER_RECIPES_TYPES: tuple[str, ...] = ("smithing_transform", "smithing_trim", "stonecutting")
UNUSED_RECIPES_TYPES: tuple[str, ...] = (
	"crafting_decorated_pot", "crafting_special_armordye", "crafting_special_bannerduplicate",
	"crafting_special_bookcloning", "crafting_special_firework_rocket", "crafting_special_firework_star",
	"crafting_special_firework_star_fade", "crafting_special_mapcloning", "crafting_special_mapextending",
	"crafting_special_repairitem", "crafting_special_shielddecoration", "crafting_special_tippedarrow",
	"crafting_transmute",
)
SPECIAL_RECIPES_TYPES: tuple[str, ...] = ("simplenergy_pulverizing", "stardust_awakened_forge")
ALL_RECIPES_TYPES: tuple[str, ...] = (*FURNACES_RECIPES_TYPES, *CRAFTING_RECIPES_TYPES, *OTHER_RECIPES_TYPES, *UNUSED_RECIPES_TYPES, *SPECIAL_RECIPES_TYPES)

# Ingr class
class Ingr(JsonDict):

	def __init__(self, id: str | JsonDict, ns: str | None = None, count: int | None = None, **kwargs: Any) -> None:
		""" Get the identity of the ingredient from its id for custom crafts

		Aliases: Ingredient(), IngrRepr()

		Args:
			id		(str):		The id of the ingredient, ex: adamantium_fragment
			ns		(str|None):	The namespace of the ingredient (optional if 'id' argument is a vanilla item), ex: iyc (default: current project id)
			count	(int|None):	The count of the ingredient (optional, used only when this ingredient format is a result item) (or use a special type of recipe that supports counts)
		Returns:
			str: The identity of the ingredient for custom crafts,
				ex: {"components":{"minecraft:custom_data":{"iyc":{"adamantium_fragment":True}}}}
				ex: {"item": "minecraft:stick"}
		Examples:
			>>> Ingr("minecraft:stick")
			{'item': 'minecraft:stick'}
			>>> Ingr("adamantium_fragment", ns="iyc")
			{'components': {'minecraft:custom_data': {'iyc': {'adamantium_fragment': True}}}}
			>>> Ingr("adamantium_fragment", ns="iyc", count=3)
			{'components': {'minecraft:custom_data': {'iyc': {'adamantium_fragment': True}}}, 'count': 3}
			>>> Ingr("diamond")
			{'components': {'minecraft:custom_data': {'detected_namespace': {'diamond': True}}}}
			>>> print(Ingr("diamond"))
			{'components': {'minecraft:custom_data': {'detected_namespace': {'diamond': True}}}}
		"""
		# Copy from another dict
		if isinstance(id, dict):
			self.update(id)
			return

		# Create from id, ns, count
		if ":" in id:
			self["item"] = id
		else:
			if ns is None:
				ns = Mem.ctx.project_id if Mem.ctx else "detected_namespace"
			self["components"] = {"minecraft:custom_data":{ns:{id:True}}}
		if count is not None:
			self["count"] = count
		self.update(kwargs)

	def copy(self) -> Ingr:
		return Ingr(dict(self))

	@stp.simple_cache
	def item_to_id(self) -> Ingr:
		""" Replace the "item" key by "id" in an item ingredient representation

		Args:
			ingr (dict): The item ingredient, ex: {"item": "minecraft:stick"}
		Returns:
			Ingr: The item ingredient representation, ex: {"id": "minecraft:stick"}

		>>> i = Ingr("minecraft:stick")
		>>> (i, i.item_to_id())
		({'item': 'minecraft:stick'}, {'id': 'minecraft:stick'})

		>>> i["Slot"] = 0

		>>> j = Ingr("adamantium_fragment", ns="iyc")
		>>> j == j.item_to_id()
		True
		"""
		if self.get("item") is None:
			return self
		if "Slot" in self:
			r: JsonDict = {"Slot": self["Slot"], "id": self["item"]}
		else:
			r: JsonDict = {"id": self["item"]}
		r.update(self)
		r.pop("item")
		return Ingr(r)

	@stp.simple_cache
	def to_id(self, add_namespace: bool = True) -> str:
		""" Get the id from an ingredient dict

		Args:
			add_namespace (bool): Whether to add the namespace to the id
		Returns:
			str: The id of the ingredient, ex: "minecraft:stick" or "iyc:adamantium_ingot"
		"""
		for k in ("item", "id"):
			if self.get(k):
				if not add_namespace and ":" in self[k]:
					return self[k].split(":")[1]
				elif add_namespace and ":" not in self[k]:
					return "minecraft:" + self[k]
				return self[k]

		custom_data: JsonDict = self["components"]["minecraft:custom_data"]
		namespace: str = ""
		id: str = ""
		for cd_ns, cd_data in custom_data.items():
			if isinstance(cd_data, dict) and cd_data:
				cd_data = cast(JsonDict, cd_data)
				first_value = next(iter(cd_data.values()))
				if isinstance(first_value, bool):
					namespace = cd_ns
					id = next(iter(cd_data.keys()))
					break
		if not namespace:
			stp.error(f"No namespace found in custom data: {custom_data}, ingredient: {self}")
		if add_namespace:
			return namespace + ":" + id
		return id

	@stp.simple_cache
	def to_name(self) -> str:
		""" Get the name of the ingredient, ex: "Stick" or "Adamantium Ingot" """
		return item_id_to_name(self.to_id(add_namespace=True))

	@stp.simple_cache
	def to_vanilla_item_id(self, add_namespace: bool = True) -> str:
		""" Get the id of the vanilla item from an ingredient dict

		Args:
			add_namespace (bool): Whether to add the namespace to the id
		Returns:
			str: The id of the vanilla item, ex: "minecraft:stick"
		"""
		ns, ingr_id = self.to_id().split(":")
		from .item import Item
		if ns == Mem.ctx.project_id:
			if add_namespace:
				return Item.from_id(ingr_id).base_item
			return Item.from_id(ingr_id).base_item.split(":")[1]
		elif ns == "minecraft":
			if add_namespace:
				return f"{ns}:{ingr_id}"
			return ingr_id
		else:
			item: str = f"{ns}:{ingr_id}"
			if Mem.external_definitions.get(item):
				if add_namespace:
					return Item.from_id(item).base_item
				return Item.from_id(item).base_item.split(":")[1]
			else:
				stp.error(f"External item '{item}' not found in the external definitions")
		return ""

	def to_item(self, id_key: str = "id") -> Ingr:
		""" Get the item data dict from an ingredient

		Args:
			id_key (str): The key to use for the item id, either "id" or "item" (default: "id")
		Returns:
			Ingr: The item data dict, ex: {"id": "minecraft:stick", "count": 1}
		"""
		ingr_id: str = self.to_id()
		ns, id = ingr_id.split(":")

		# Minecraft item
		if ns == "minecraft":
			return Ingr({id_key: id, "count": 1})

		# Get from internal definitions
		from .item import Item
		if ns == Mem.ctx.project_id:
			item_data = Item.from_id(id)
			result = Ingr({id_key: item_data.base_item, "count": 1})

			# Add components
			for k, v in item_data.components.items():
				if result.get("components") is None:
					result["components"] = {}
				if k.startswith("!"):
					result["components"][f"!minecraft:{k[1:]}"] = {}
				else:
					result["components"][f"minecraft:{k}"] = v
			return result

		# External definitions
		if Mem.external_definitions.get(ingr_id):
			item_data = Item.from_id(ingr_id)
			result = Ingr({id_key: item_data.base_item, "count": 1})

			# Add components
			for k, v in item_data.components.items():
				if result.get("components") is None:
					result["components"] = {}
				if k.startswith("!"):
					result["components"][f"!minecraft:{k[1:]}"] = {}
				else:
					result["components"][f"minecraft:{k}"] = v
			return result

		stp.error(f"External item '{ingr_id}' not found in the external definitions")
		return Ingr({})

	@stp.simple_cache
	def register_loot_table(self, result_count: int | JsonDict) -> str:
		""" Get the loot table for an ingredient dict

		Args:
			result_count (int|dict): The count of the result item, can be an int or a dict for random counts
				ex: 1
				ex: {"type": "minecraft:uniform","min": 4,"max": 6}
		Returns:
			str: The loot table path, ex: "my_datapack:i/stick"
		"""
		# If item from this datapack
		item: str = self.to_id()
		if item.startswith(Mem.ctx.project_id):
			item = item.split(":")[1]
			loot_table = f"{Mem.ctx.project_id}:i/{item}{result_count_to_suffix(result_count)}"
			return loot_table

		namespace, item = item.split(":")
		loot_table = f"{Mem.ctx.project_id}:recipes/{namespace}/{item}{result_count_to_suffix(result_count)}"

		# If item from another datapack, generate the loot table
		if namespace != "minecraft":
			file: JsonDict = {"pools":[{"rolls":1,"entries":[{"type":"minecraft:loot_table","value": f"{Mem.ctx.project_id}:external/{namespace}/{item}"}] }] }
		else:
			file: JsonDict = {"pools":[{"rolls":1,"entries":[{"type":"minecraft:item","name":f"{namespace}:{item}"}] }] }
		if (isinstance(result_count, int) and result_count > 1) or hasattr(result_count, "get"):
			file["pools"][0]["entries"][0]["functions"] = [{"function": "minecraft:set_count","count": result_count}]

		Mem.ctx.data[loot_table] = set_json_encoder(LootTable(file), max_level=9)
		return loot_table

	@staticmethod
	@stp.simple_cache
	def get_ingredients_from_vanilla_recipe(recipe: JsonDict) -> list[str]:
		""" Get the ingredients from a vanilla recipe dict

		Args:
			recipe (dict): The final recipe JSON dict, ex:

			{
				"type": "minecraft:crafting_shaped",
				"pattern": [...],
				"key": {...},
				"result": {...}
			}
		Returns:
			list[str]: The ingredients ids
		"""
		if recipe.get("key"):
			return list(recipe["key"].values())
		elif recipe.get("ingredients"):
			return recipe["ingredients"]
		elif recipe.get("ingredient"):
			return [recipe["ingredient"]]
		elif recipe.get("template"):
			return [recipe["template"]]
		else:
			return []

# Type aliases
IngrRepr = Ingredient = Ingr




from .cd_utils import get_supported_versions as get_supported_versions
from beet.core.utils import JsonDict as JsonDict

SMITHED_API_URL: str

def validate_credentials(credentials: JsonDict) -> tuple[str, str]:
    """ Get and validate Smithed credentials

\tArgs:
\t\tcredentials (dict[str, str]): Credentials for the Smithed API
\tReturns:
\t\tstr: API key for Smithed
\t\tstr: GitHub author
\t"""
def validate_config(smithed_config: dict[str, str]) -> tuple[str, str, str]:
    """ Validate Smithed configuration

\tArgs:
\t\tsmithed_config (dict[str, str]): Configuration for the Smithed project
\tReturns:
\t\tstr: Project name on Smithed
\t\tstr: Version of the project
\t\tstr: Slug (namespace) of the project
\t"""
def upload_version(project_id: str, project_name: str, version: str, api_key: str, author: str) -> None:
    """ Upload new version

\tArgs:
\t\tproject_id\t\t(str):\t\t\t\tSmithed project ID
\t\tproject_name\t(str):\t\t\t\tName of the project
\t\tversion\t\t\t(str):\t\t\t\tVersion number
\t\tapi_key\t\t\t(str):\t\t\t\tAPI key for Smithed
\t\tauthor\t\t\t(str):\t\t\t\tAuthor (for the github link)
\t"""
def upload_to_smithed(credentials: dict[str, str], smithed_config: dict[str, str], changelog: str = '') -> None:
    """ Upload the project to Smithed using the credentials and the configuration

\tArgs:
\t\tcredentials\t\t(dict[str, str]):\tCredentials for the Smithed API
\t\tsmithed_config\t(dict[str, str]):\tConfiguration for the Smithed project
\t\tchangelog\t\t(str):\t\t\t\tChangelog text for the release
\t"""




# Imports
import stouputils as stp
from beet import Advancement, Recipe
from beet.core.utils import JsonDict

from ...core.__memory__ import Mem
from ...core.cls.ingredients import Ingr
from ...core.cls.item import Item
from ...core.cls.recipe import (
    BlastingRecipe,
    CampfireCookingRecipe,
    CraftingShapedRecipe,
    CraftingShapelessRecipe,
    SmeltingRecipe,
    SmithingTransformRecipe,
    SmithingTrimRecipe,
    SmokingRecipe,
    StonecuttingRecipe,
)
from ...core.utils.io import set_json_encoder, write_function


class VanillaRecipeHandler:
    """ Handler for vanilla recipe generation.

    This class handles the generation of vanilla recipes (shapeless, shaped, furnace, ...).
    """

    def __init__(self) -> None:
        """ Initialize the handler. """
        self.vanilla_generated_recipes: list[tuple[str, str]] = []

    @classmethod
    def routine(cls) -> None:
        """ Main routine for vanilla recipe generation. """
        handler = cls()
        handler.generate_recipes()

        # Create recipe unlocking function if vanilla recipes were generated
        if handler.vanilla_generated_recipes:
            # Create a function that will give all recipes
            content = "\n# Get all recipes\n"
            for recipe_file, _ in handler.vanilla_generated_recipes:
                content += f"recipe give @s {Mem.ctx.project_id}:{recipe_file}\n"
            write_function(f"{Mem.ctx.project_id}:utils/get_all_recipes", content + "\n")

            # Get all ingredients and their associated recipes
            ingredients: dict[str, set[str]] = {}
            for recipe_name, _ in handler.vanilla_generated_recipes:
                recipe: JsonDict = Mem.ctx.data[Mem.ctx.project_id].recipes[recipe_name].data
                for ingr_str in Ingr.get_ingredients_from_vanilla_recipe(recipe):
                    if ingr_str not in ingredients:
                        ingredients[ingr_str] = set()
                    ingredients[ingr_str].add(recipe_name)

            # Write the advancement
            adv_path: str = f"{Mem.ctx.project_id}:unlock_recipes"
            adv_json: JsonDict = {
                "criteria": {"requirement": {"trigger": "minecraft:inventory_changed"}},
                "rewards": {"function": f"{Mem.ctx.project_id}:advancements/unlock_recipes"}
            }
            Mem.ctx.data[adv_path] = set_json_encoder(Advancement(adv_json), max_level=-1)

            # Write the function that will unlock the recipes
            content = f"""
# Revoke advancement
advancement revoke @s only {Mem.ctx.project_id}:unlock_recipes

## For each ingredient in inventory, unlock the recipes
"""
            # Add ingredients
            for ingr, recipes in ingredients.items():
                recipes_list: list[str] = sorted(recipes)
                content += (
                    f"# {ingr}\nscoreboard players set #success {Mem.ctx.project_id}.data 0\n"
                    f"execute store success score #success {Mem.ctx.project_id}.data if items entity @s container.* {ingr}\n"
                )
                for recipe_path in recipes_list:
                    content += f"execute if score #success {Mem.ctx.project_id}.data matches 1 run recipe give @s {Mem.ctx.project_id}:{recipe_path}\n"
                content += "\n"

            # Add result items
            content += "## Add result items\n"
            for recipe_name, item in handler.vanilla_generated_recipes:
                content += f"""execute if items entity @s container.* *[custom_data~{{"{Mem.ctx.project_id}": {{"{item}":true}} }}] run recipe give @s {Mem.ctx.project_id}:{recipe_name}\n"""

            write_function(f"{Mem.ctx.project_id}:advancements/unlock_recipes", content)

    def vanilla_shapeless_recipe(self, recipe: CraftingShapelessRecipe, item: str) -> JsonDict:
        """Generate a vanilla shapeless recipe.

        Args:
            recipe (CraftingShapelessRecipe): The recipe data.
            item (str): The item to generate the recipe for.

        Returns:
            JsonDict: The generated recipe.
        """
        result_ingr = recipe.result or Ingr(item)
        ingredients: list[str] = [i.to_vanilla_item_id() for i in recipe.ingredients]

        to_return: JsonDict = {
            "type": "minecraft:" + recipe.type,
            "category": recipe.category,
            "group": recipe.group,
            "ingredients": ingredients,
            "result": result_ingr.to_item().item_to_id(),
        }

        if not to_return["group"]:
            del to_return["group"]

        to_return["result"]["count"] = recipe["result_count"]
        return to_return

    def vanilla_shaped_recipe(self, recipe: CraftingShapedRecipe, item: str) -> JsonDict:
        """Generate a vanilla shaped recipe.

        Args:
            recipe (CraftingShapedRecipe): The recipe data.
            item (str): The item to generate the recipe for.

        Returns:
            JsonDict: The generated recipe.
        """
        result_ingr = recipe.result or Ingr(item, Mem.ctx.project_id)
        ingredients: dict[str, str] = {
            k: i.to_vanilla_item_id()
            for k, i in recipe.ingredients.items()
        }

        to_return: JsonDict = {
            "type": "minecraft:" + recipe.type,
            "category": recipe.category,
            "group": recipe.group,
            "pattern": recipe.shape,
            "key": ingredients,
            "result": result_ingr.to_item().item_to_id(),
        }

        if not to_return["group"]:
            del to_return["group"]

        to_return["result"]["count"] = recipe.result_count
        return to_return

    @stp.simple_cache
    def vanilla_furnace_recipe(self, recipe: SmeltingRecipe | BlastingRecipe | SmokingRecipe | CampfireCookingRecipe, item: str) -> JsonDict:
        """ Generate a vanilla furnace recipe.

        Args:
            recipe (SmeltingRecipe | BlastingRecipe | SmokingRecipe | CampfireCookingRecipe): The recipe data.
            item (str): The item to generate the recipe for.

        Returns:
            JsonDict: The generated recipe.
        """
        result_ingr = recipe.result or Ingr(item)
        ingredient_vanilla: str = recipe.ingredient.to_vanilla_item_id()

        to_return: JsonDict = {
            "type": "minecraft:" + recipe.type,
            "category": recipe.category,
            "group": recipe.group,
            "ingredient": ingredient_vanilla,
            "result": result_ingr.to_item().item_to_id(),
        }

        if not to_return["group"]:
            del to_return["group"]

        to_return["result"]["count"] = recipe.result_count
        return to_return

    @stp.simple_cache
    def vanilla_stonecutting_recipe(self, recipe: StonecuttingRecipe, item: str) -> JsonDict:
        """ Generate a vanilla stonecutting recipe.

        Args:
            recipe (StonecuttingRecipe): The recipe data.
            item (str): The item to generate the recipe for.

        Returns:
            JsonDict: The generated recipe.
        """
        result_ingr = recipe.result or Ingr(item, Mem.ctx.project_id)
        ingredient_vanilla: str = recipe.ingredient.to_vanilla_item_id()

        to_return: JsonDict = {
            "type": "minecraft:" + recipe.type,
            "group": recipe.group,
            "ingredient": ingredient_vanilla,
            "result": result_ingr.to_item().item_to_id(),
        }

        if not to_return["group"]:
            del to_return["group"]

        to_return["result"]["count"] = recipe.result_count
        return to_return

    @stp.simple_cache
    def vanilla_smithing_transform_recipe(self, recipe: SmithingTransformRecipe, item: str) -> JsonDict:
        """ Generate a vanilla smithing transform recipe.

        Args:
            recipe (SmithingTransformRecipe): The recipe data.
            item (str): The item to generate the recipe for.

        Returns:
            JsonDict: The generated recipe.
        """
        result_ingr = recipe.result or Ingr(item)
        to_return: JsonDict = {
            "type": "minecraft:" + recipe.type,
            "base": recipe.base.to_vanilla_item_id(),
            "addition": recipe.addition.to_vanilla_item_id(),
            "template": recipe.template.to_vanilla_item_id(),
            "result": result_ingr.to_item().item_to_id(),
        }
        to_return["result"]["count"] = recipe.result_count
        return to_return

    @stp.simple_cache
    def vanilla_smithing_trim_recipe(self, recipe: SmithingTrimRecipe, item: str) -> JsonDict:
        """ Generate a vanilla smithing trim recipe.

        Args:
            recipe (SmithingTrimRecipe): The recipe data.
            item (str): The item to generate the recipe for.

        Returns:
            JsonDict: The generated recipe.
        """
        return {
            "type": "minecraft:" + recipe.type,
            "base": recipe.base.to_vanilla_item_id(),
            "addition": recipe.addition.to_vanilla_item_id(),
            "template": recipe.template.to_vanilla_item_id(),
            "pattern": recipe.pattern,
        }

    def generate_recipes(self, override: list[str] | None = None) -> None:
        """ Generate all vanilla recipes.

        Args:
            override (list[str]): If set, only generate recipes for this item and with override only. Used to regenerate a specific item.
        """
        for item in Mem.definitions.keys():
            if override and item not in override:
                continue
            obj = Item.from_id(item)

            i = 1
            for recipe in obj.recipes:
                name = f"{item}" if i == 1 else f"{item}_{i}"

                # Handle different recipe types
                if recipe["type"] == CraftingShapelessRecipe.type:
                    recipe = CraftingShapelessRecipe.from_dict(recipe)
                    if all(i.get("item") for i in recipe.ingredients):
                        self.write_recipe_file(name, self.vanilla_shapeless_recipe(recipe, item))
                        i += 1
                        self.vanilla_generated_recipes.append((name, item))

                elif recipe["type"] == CraftingShapedRecipe.type:
                    recipe = CraftingShapedRecipe.from_dict(recipe)
                    if all(i.get("item") for i in recipe.ingredients.values()):
                        self.write_recipe_file(name, self.vanilla_shaped_recipe(recipe, item))
                        i += 1
                        self.vanilla_generated_recipes.append((name, item))

                elif recipe["type"] in (SmeltingRecipe.type, BlastingRecipe.type, SmokingRecipe.type, CampfireCookingRecipe.type):
                    recipe = SmeltingRecipe.from_dict(recipe) if recipe["type"] == SmeltingRecipe.type else \
                             BlastingRecipe.from_dict(recipe) if recipe["type"] == BlastingRecipe.type else \
                             SmokingRecipe.from_dict(recipe) if recipe["type"] == SmokingRecipe.type else \
                             CampfireCookingRecipe.from_dict(recipe)
                    if recipe.ingredient.get("item"):
                        self.write_recipe_file(name, self.vanilla_furnace_recipe(recipe, item))
                        i += 1
                        self.vanilla_generated_recipes.append((name, item))

                elif recipe["type"] == StonecuttingRecipe.type:
                    recipe = StonecuttingRecipe.from_dict(recipe)
                    if recipe.ingredient.get("item"):
                        self.write_recipe_file(name, self.vanilla_stonecutting_recipe(recipe, item))
                        i += 1
                        self.vanilla_generated_recipes.append((name, item))

                elif recipe["type"] == SmithingTransformRecipe.type:
                    recipe = SmithingTransformRecipe.from_dict(recipe)
                    if (recipe.base.get("item") and recipe.addition.get("item") and recipe.template.get("item")):
                        self.write_recipe_file(name, self.vanilla_smithing_transform_recipe(recipe, item))
                        i += 1
                        self.vanilla_generated_recipes.append((name, item))

                elif recipe["type"] == SmithingTrimRecipe.type:
                    recipe = SmithingTrimRecipe.from_dict(recipe)
                    if (recipe.base.get("item") and recipe.addition.get("item") and recipe.template.get("item") and recipe.pattern):
                        self.write_recipe_file(name, self.vanilla_smithing_trim_recipe(recipe, item))
                        i += 1
                        self.vanilla_generated_recipes.append((name, item))

    def write_recipe_file(self, name: str, content: JsonDict) -> None:
        """ Write a recipe file.

        Args:
            name (str): The name of the recipe.
            content (JsonDict): The recipe content.
        """
        Mem.ctx.data[Mem.ctx.project_id].recipes[name] = set_json_encoder(Recipe(content), max_level=-1)




import stouputils as stp
from .cd_utils import get_supported_versions as get_supported_versions
from beet.core.utils import JsonDict as JsonDict

MODRINTH_API_URL: str
PROJECT_ENDPOINT: str
VERSION_ENDPOINT: str

def validate_credentials(credentials: dict[str, str]) -> str:
    """ Get and validate Modrinth credentials

\tArgs:
\t\tcredentials (dict[str, str]): Credentials for the Modrinth API
\tReturns:
\t\tstr: API key for Modrinth
\t"""
def validate_config(modrinth_config: JsonDict) -> tuple[str, str, str, str, str, str, str]:
    """ Validate Modrinth configuration

\tArgs:
\t\tmodrinth_config (JsonDict): Configuration for the Modrinth project
\tReturns:
\t\tstr: Project name on Modrinth
\t\tstr: Version of the project
\t\tstr: Slug (namespace) of the project
\t\tstr: Summary of the project
\t\tstr: Description in Markdown format
\t\tstr: Version type (release, beta, alpha)
\t\tstr: Build folder path
\t"""
def get_project(slug: str, headers: dict[str, str]) -> dict[str, str]:
    """ Get project from Modrinth

\tArgs:
\t\tslug (str): Project slug/namespace
\t\theaders (dict[str, str]): Headers for Modrinth API requests
\tReturns:
\t\tdict: Project data
\t"""
def update_project_description(slug: str, description: str, summary: str, headers: dict[str, str]) -> None:
    """ Update project description and summary

\tArgs:
\t\tslug (str): Project slug/namespace
\t\tdescription (str): Project description in Markdown
\t\tsummary (str): Project summary
\t\theaders (dict[str, str]): Headers for Modrinth API requests
\t"""
def handle_existing_version(slug: str, version: str, headers: dict[str, str]) -> bool:
    """ Check and handle existing version

\tArgs:
\t\tslug (str): Project slug/namespace
\t\tversion (str): Version to check
\t\theaders (dict[str, str]): Headers for Modrinth API requests
\tReturns:
\t\tbool: True if we should continue, False otherwise
\t"""
def generate_fabric_metadata(mod_id: str, metadata: JsonDict) -> str:
    """ Generate Fabric mod metadata JSON

\tArgs:
\t\tmod_id (str): Mod ID
\t\tmetadata (dict): Mod metadata
\tReturns:
\t\tstr: Fabric mod.json content
\t"""
def generate_forge_metadata(mod_id: str, metadata: JsonDict, is_neoforge: bool = False) -> str:
    """ Generate Forge/NeoForge mod metadata TOML

\tArgs:
\t\tmod_id (str): Mod ID
\t\tmetadata (dict): Mod metadata
\t\tis_neoforge (bool): Whether this is for NeoForge (uses javafml) or Forge (uses lowcodefml)
\tReturns:
\t\tstr: mods.toml content
\t"""
def generate_quilt_metadata(mod_id: str, metadata: JsonDict) -> str:
    """ Generate Quilt mod metadata JSON

\tArgs:
\t\tmod_id (str): Mod ID
\t\tmetadata (dict): Mod metadata
\tReturns:
\t\tstr: quilt.mod.json content
\t"""
def convert_datapack_to_mod(datapack_path: str, output_path: str, metadata: JsonDict, platforms: list[str], resource_pack_path: str | None = None) -> None:
    """ Convert a datapack ZIP to a mod JAR with proper metadata files

\tArgs:
\t\tdatapack_path (str): Path to the datapack ZIP file
\t\toutput_path (str): Path where to save the mod JAR file
\t\tmetadata (dict): Mod metadata (id, name, version, description, authors, etc.)
\t\tplatforms (list[str]): List of platforms (fabric, forge, neoforge, quilt)
\t\tresource_pack_path (str): Optional path to the resource pack ZIP file
\t"""
def get_file_parts(project_name: str, build_folder: str, modrinth_config: JsonDict, project_data: JsonDict | None = None) -> list[str]:
    """ Get file parts to upload

\tArgs:
\t\tproject_name (str): Name of the project
\t\tbuild_folder (str): Path to build folder
\t\tmodrinth_config (dict): Modrinth configuration
\t\tproject_data (dict): Optional project data from Modrinth API
\tReturns:
\t\tlist[str]: List of file paths to upload
\t"""
def upload_version(project_id: str, project_name: str, version: str, version_type: str, changelog: str, file_parts: list[str], headers: dict[str, str], dependencies: list[str] | None = None, loaders: list[str] | None = None) -> JsonDict:
    """ Upload new version

\tArgs:
\t\tproject_id\t\t(str):\t\t\t\tModrinth project ID
\t\tproject_name\t(str):\t\t\t\tName of the project
\t\tversion\t\t\t(str):\t\t\t\tVersion number
\t\tversion_type\t(str):\t\t\t\tType of version (release, beta, alpha)
\t\tchangelog\t\t(str):\t\t\t\tChangelog text
\t\tfile_parts\t\t(list[str]):\t\tList of files to upload
\t\theaders\t\t\t(dict[str, str]):\tHeaders for Modrinth API requests
\t\tdependencies\t(list[str]):\t\tList of dependencies
\t\tloaders\t\t\t(list[str]):\t\tList of loaders (datapack, fabric, forge, etc.)
\tReturns:
\t\tdict: Upload response data
\t"""
def set_resource_pack_required(version_id: str, resource_pack_hash: str, headers: dict[str, str]) -> None:
    """ Set resource pack as required

\tArgs:
\t\tversion_id (str): ID of the version
\t\tresource_pack_hash (str): SHA1 hash of resource pack
\t\theaders (dict[str, str]): Headers for Modrinth API requests
\t"""
@stp.handle_error
def upload_to_modrinth(credentials: dict[str, str], modrinth_config: JsonDict, changelog: str = '') -> None:
    """ Upload the project to Modrinth using the credentials and the configuration

\tArgs:
\t\tcredentials\t\t(dict[str, str]):\tCredentials for the Modrinth API
\t\tmodrinth_config\t(dict):\t\t\t\tConfiguration for the Modrinth project
\t\tchangelog\t\t(str):\t\t\t\tChangelog text for the release
\t"""
